/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */

/**
 *  \ Benchmark for collision detection \
 */

/*!
 *  @file 
 *  @brief Benchmark program for collision detection algorithms. Type
 *  './bench -h' for an explanation of the command line options and keys.
 *  Keys: (interactive version)
 *       - l : lighting on/off
 *       - p : polygon mode (wireframe, dots, filled)
 *       - space : pauseBench
 *       - s : show closeste points of moving object ( works not when using offset-objects! )
 *       - a : stop/restart configuration computation when finished
 *       - m : grab screenshots in ./movie-directory
 *       - t : slow motion in movie
 *       - u : turns on speed up in movie recording mode
 *       - i : increase speed up in movie recording mode 
 *  @author Sven Trenkel, Rene Weller, Gabriel Zachmann
 */

/*
 * Includes ;
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <algorithm>
#include <list>
#include <vector>

#define OSG_WITH_GLUT

#include <OpenSG/OSGGLUT.h>
#include <OpenSG/OSGConfig.h>

//#include <GL/glut.h>

#include <OpenSG/OSGBaseTypes.h>
#include <OpenSG/OSGBaseFunctions.h>
#include <OpenSG/OSGGeometry.h>
#include <OpenSG/OSGGeoProperty.h>
#include <OpenSG/OSGGeoFunctions.h>
#include <OpenSG/OSGVector.h>
#include <OpenSG/OSGQuaternion.h>
#include <OpenSG/OSGMatrix.h>
#include <OpenSG/OSGTransform.h>
#include <OpenSG/OSGCamera.h>
#include <OpenSG/OSGPerspectiveCamera.h>
#include <OpenSG/OSGDrawAction.h>
#include <OpenSG/OSGSolidBackground.h>
#include <OpenSG/OSGGroup.h>
#include <OpenSG/OSGDirectionalLight.h>
#include <OpenSG/OSGViewport.h>
#include <OpenSG/OSGBoxVolume.h>
#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGSimpleAttachments.h>
#include <OpenSG/OSGSimpleSceneManager.h>
#include <OpenSG/OSGAttachment.h>
#include <OpenSG/OSGGLUTWindow.h>
#include <OpenSG/OSGTrackball.h>
#include <OpenSG/OSGFaceIterator.h>
#include <OpenSG/OSGSimpleMaterial.h>
#include <OpenSG/OSGSceneFileHandler.h>
#include <OpenSG/OSGTriangleIterator.h>
#include <OpenSG/OSGMaterialGroup.h>
#include <OpenSG/OSGFileGrabForeground.h>

#include "Bench.h"
#include "Settings.h"
#include "Grid.h"
#include "Utils.h"

#ifdef _WIN32
#include <windows.h>
#ifdef max
#undef max
#endif
#ifdef min
#undef min
#endif
#else
#include <sys/time.h>
#include <sys/resource.h>
#include <dirent.h>
#include <dlfcn.h>
#include <locale.h>
#endif
using osg:: beginEditCP;
using osg:: endEditCP;
using osg:: Vec3f;

/**
 *  \ Global variables \
 */
osg::SimpleSceneManager * mgr;
PosFinder                           *posFinder;         /* class for finding
                                                         * the positions to
                                                         * benchmark */
Algorithm                           *algorithm = 0;     /* indicates the chosen
                                                         * algorithm */
Settings                            *settings;          /* class for
                                                         * benchmarking
                                                         * parameters */
static bool                         pauseBench = true;  /* pauseBenchs */
static bool                         showPos  = false;   /* show closest points */
static bool                         showing = false;   /* display benchmark,
                                                         * timings will be
                                                         * bogus */
static bool                         finishedWinMode = false;   /* if true:
                                                                 * stops after
                                                                 * all
                                                                 * iterations
                                                                 * are
                                                                 * finished,
                                                                 * else restart
                                                                 * benchmark */
static bool                         grabScreenOn = false;      /* turns on
                                                                 * screen
                                                                 * grabbing */
static bool                         slowMo = false;            /* slow motion
                                                                 * during
                                                                 * filming */
static bool                         speedUp = false;           /* drops some
                                                                 * iterations
                                                                 * during
                                                                 * filming */
static unsigned int                 slowMoFrames = 8;           /* nr of frames
                                                                 * between two
                                                                 * iterations
                                                                 * in slow-mo
                                                                 * mode */
static unsigned int                 slowMoCounter = 0;          /* counter for
                                                                 * the number
                                                                 * of frames */
static unsigned int                 speedUpFrames = 25;         /* nr of
                                                                 * iterations
                                                                 * between two
                                                                 * frames in
                                                                 * speedUp-mode
                                                                 * */
static unsigned int                 speedUpCounter = 0;         /* counter for
                                                                 * nr of
                                                                 * iterations */
static unsigned int                 endSequenceFrames = 100;    /* nr of frames
                                                                 * displayed
                                                                 * after last
                                                                 * iteration
                                                                 * (for
                                                                 * filming) if
                                                                 * in
                                                                 * finishedWinMode
                                                                 * */
static unsigned int                 endSequenceCounter = 0;     /* counter for
                                                                 * the number
                                                                 * of
                                                                 * end-frames */
static osg::NodePtr                 scene;      /* root of the scene */
static osg::NodePtr                 movingNode; /* node for the moving object */
static osg::NodePtr                 fixedNode;  /* node for the fixed object */
static osg::NodePtr                 posNode;    /* father node of the dotNodes
                                                 * that remark the closest
                                                 * point */
static osg::NodePtr                 distNode;   /* Offset-surface if not using PQP */
static osg::SimpleMaterialPtr       distNodeMat;/* material for the closest
                                                 * point */
static osg::SimpleMaterialPtr       posNodeMat; /* material for the closest
                                                 * point */
static osg::GeometryPtr             dotNode( osg::NullFC ); /* ball that
                                                             * indicates
                                                             * closest point
                                                             * between moving-
                                                             * and fixed object
                                                             * */
static osg::FileGrabForegroundPtr   fileGrab;           /* display-grabber for
                                                         * filming */

/**
 *  \ Forward Declarations \
 */
void                                grabScreen();       /* fct for screen
                                                         * grabbing */
void                                grabScreenInit();   /* inits screen grabbing */

/* fcts for loading external collision detection libraries */
#ifdef _WIN32
typedef HINSTANCE                   Mod;
#define openMod     LoadLibrary
#define closeMod    FreeLibrary
#define dlerror()   "Error loading module"
#else
typedef void                        *Mod;
#define openMod( x )    dlopen( (x), RTLD_NOW )
#define GetProcAddress  dlsym
#define closeMod        dlclose
#endif
#ifdef _WIN32
typedef std::vector<Algorithm *> *initModule(void);
#else
typedef std::vector<Algorithm *> *initModule(void);
#endif


/*! ----------------------------------------------------------------------------
 *  \brief  Used to register an algorithm. Modules call this function in order
 *          to register their algorithm(s) with the program. After doing so it
 *          can be selected with the -A parameter and is listed in the -h help
 *          notes.
 *  \param Shortname The two or three letter name used with the
 *          -A parameter. Must be unique.
 *  \param name The full name of the
 *          algorithm.
 *  \param api The api struct used to access the algorithm.
 *  \return 1 on error, 0 otherwise.
 ------------------------------------------------------------------------------- */
int registerAlgorithm( const char *shortName, const char *name, tAPI *api )
{
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~*/
    Algorithm   *a, *prev = 0;
    /*~~~~~~~~~~~~~~~~~~~~~~~~~~*/

    for( a = algorithm; a; a = a->next )
    {
        if( !strcmp( a->shortName, shortName ) )
        {
            printf( "Unable to register algorithm %s (%s): Already registered %s (%s)\n",
                name, shortName, a->name, a->shortName );
            return 1;
        }

        prev = a;
    }

    a = new Algorithm;
    if( prev )
    {
        prev->next = a;
    }
    else
    {
        algorithm = a;
    }

    a->shortName = shortName;
    a->name = name;
    a->api = *api;
    a->next = 0;
    return 0;
}

/*! ----------------------------------------------------------------------------
 *  \brief  Load a module. Loads a module and calls it's init function
 *          "IniModule" (C style). 
 *  \param filename Full filename of the module
 *          to load.
 ------------------------------------------------------------------------------- */
void loadModule( const char *filename )
{
    Mod         module = openMod( filename );
    if( !module )
    {
        fprintf( stderr, "%s\n", dlerror() );
        return;
    }
    
    initModule   *init;
#ifdef _WIN32
    init = ( initModule * ) GetProcAddress( module, "initModule" );    
#else
    init = ( initModule * ) GetProcAddress( module, "initModule" );    
#endif
  //  initModule   *init = ( initModule * ) GetProcAddress( module, "initModule" );
    
    if( !init )
    {
        fprintf( stderr, "%s\n", dlerror() );
        return;
    }

    std::vector<Algorithm *> *ret = init();
    if( !ret )
    {
        fprintf( stderr, "%s: ModuleInit Fehler\n", filename );
        return;
    }
    
    else
    {
        for( std::vector<Algorithm *>::iterator it = ret->begin(); it != ret->end(); it++ )
        {
            registerAlgorithm( (*it)->shortName, (*it)->shortName, &(*it)->api );
        }
    }
}

/*! ----------------------------------------------------------------------------
 *  \brief  Load all modules Calls loadModule() for every file ending with
 *          ".dll" or ".so" in a certain directory. 
 *  \param name name of the directory to search.
 ------------------------------------------------------------------------------- */
void loadDir( const char *name )
{
#ifdef _WIN32
    WIN32_FIND_DATA dir;
    HANDLE          fhandle;
    char            directory[256];

    /* better: snprintf() */
    sprintf( directory, "%s\\*.*", name );

    /* Handle to directory */
    if( (fhandle = FindFirstFile( directory, &dir )) != INVALID_HANDLE_VALUE )
    {
        do
        {   /* read directory
             * ;
             * printf("%s\n",
             * dir.cFileName);
             * */
            int len = strlen( dir.cFileName );
            if( len <= 3 || (strcmp( dir.cFileName + (len - 3), ".so" )
            && strcmp( dir.cFileName + (len - 4), ".dll" )) )
                continue;

            /*
             * strcpy( fullname + namelen, de->d_name );
             */
            loadModule( dir.cFileName );
        } while( FindNextFile( fhandle, &dir ) );
    }

    FindClose( fhandle );

#else
    char            *fullname;
    int             namelen = strlen( name ), len;
    DIR             *dir;
    struct dirent   *de = 0;

    dir = opendir( name );
    if( !dir )
    {
        fprintf( stderr, "Could not open module directory %s\n", name );
        return;
    }

    fullname = new char[namelen + 1 + sizeof( de->d_name )];
    strcpy( fullname, name );
    fullname[namelen++] = '/';
    while( (de = readdir( dir )) )
    {
        len = strlen( de->d_name );
        if( len <= 3 || (strcmp( de->d_name + (len - 3), ".so" )
        && strcmp( de->d_name + (len - 4), ".dll" )) )
            continue;
        strcpy( fullname + namelen, de->d_name );
        loadModule( fullname );
    }

    delete[] fullname;
#endif
}

/*! ----------------------------------------------------------------------------
 *  Do a benchmark for a single configuration in windows-mode
 ------------------------------------------------------------------------------- */
static void iterate()
{
    bool    ret = false;
    
    if( dotNode == osg::NullFC )
    {
        dotNode = osg::makeSphereGeo( 1, 0.0075 );
        addRefCP( dotNode );
        addRefCP( movingNode );

        posNode = osg::Node::create();
        addRefCP( posNode );
        scene = posFinder->getScene();
        
        beginEditCP( dotNode );
        dotNode->setMaterial( posNodeMat );
        endEditCP( dotNode );

        beginEditCP( posNode );
        posNode->setCore( osg::Group::create() );
        endEditCP( posNode );
    }

    if( !finishedWinMode )
    {
        ret = posFinder->iterate();
    }

    if ( settings->searchAlgorithm->api.getDistance )
    {
        for( int i = 0;  i < posFinder->testQueue.size(); ++i )
        {
            osg::Pnt3f          pos;
            osg::Matrix matrix( posFinder->testQueue[i] );
            double              d = settings->searchAlgorithm->api.getDistance( 1, &matrix,
                                                                                &pos, NULL );
            osg::NodePtr        transNode = osg::Node::create(),
            geoNode = osg::Node::create();

            osg::TransformPtr   trans = osg::Transform::create();
            beginEditCP( posNode );

            matrix.multMatrixPnt( pos );
            matrix.setIdentity();
            matrix.setTranslate( pos );
            trans->getSFMatrix()->setValue( matrix );

            beginEditCP( geoNode );
            geoNode->setCore( dotNode );
            endEditCP( geoNode );

            beginEditCP( transNode );
            transNode->setCore( trans );
            transNode->addChild( geoNode );
            endEditCP( transNode );

            posNode->addChild( transNode );
            endEditCP( posNode );
        }
    }
    
    else
    {
        for( int i = 0;  i < posFinder->testQueue.size(); ++i )
        {
            osg::Pnt3f          pos;
            osg::Matrix matrix( posFinder->testQueue[i] );
            
            osg::NodePtr        transNode = osg::Node::create(),
            geoNode = osg::Node::create();

            osg::TransformPtr   trans = osg::Transform::create();
            beginEditCP( posNode );

            //pos = matrix.getTranslate();
            //matrix.multMatrixPnt( pos );
            pos[0] = matrix[3][0];
            pos[1] = matrix[3][1];
            pos[3] = matrix[3][2];
            
            //matrix.multMatrixPnt( pos );
            matrix.setIdentity();
            matrix.setTranslate( pos );
            trans->getSFMatrix()->setValue( matrix );

            beginEditCP( geoNode );
            geoNode->setCore( dotNode );
            endEditCP( geoNode );

            beginEditCP( transNode );
            transNode->setCore( trans );
            transNode->addChild( geoNode );
            endEditCP( transNode );

            posNode->addChild( transNode );
            endEditCP( posNode );
        }
    }
    
    posFinder->testQueue.resize( 0 );

    if( ret )
    {
        if( posFinder->rotate() )
            pauseBench = true;
    }

    if( speedUp && speedUpCounter < speedUpFrames )
    {
        speedUpCounter++;
        iterate();
    }
    else if( speedUp )
        speedUpCounter = 0;
}

/** ----------------------------------------------------------------------------
 *  \ Display and User input/ only active in windows-mode \
 ------------------------------------------------------------------------------- */
void display()
{
    if( (!pauseBench && !slowMo) || (!pauseBench && slowMo && (slowMoCounter >= slowMoFrames)) )
    {
        iterate();
        if( slowMo )
            slowMoCounter = 0;
    }

    if( slowMo && !pauseBench )
        slowMoCounter++;

    mgr->redraw();

    if( grabScreenOn )
        grabScreen();

    if( finishedWinMode )
    {
        if( endSequenceCounter < endSequenceFrames )
        {
            endSequenceCounter++;
        }
        else
        {
            delete mgr;
            osg::osgExit();
            exit( 0 );
        }
    }
}

void reshape( int w, int h )
{
    mgr->resize( w, h );
}

void animate()
{
    glutPostRedisplay();
}

void motion( int x, int y )
{
    mgr->mouseMove( x, y );
    return;
}

void mouse( int button, int state, int x, int y )
{
    if( state == GLUT_UP )
        mgr->mouseButtonRelease( button, x, y );
    else
        mgr->mouseButtonPress( button, x, y );
}

void vis( int visible )
{
    if( visible == GLUT_VISIBLE )
        glutIdleFunc( animate );
    else
        glutIdleFunc( NULL );
}

/* -----------------------------------------------------------------------------
 * \brief process Keyboard input
 *  Keys:
 *       - l : lighting on/off
 *       - p : polygon mode (wireframe, dots, filled)
 *       - space : pauseBench
 *       - s : show closeste points of moving object ( works not when using offset-surfaces! )
 *       - a : stop/restart configuration computation when finished
 *       - m : grab screenshots in ./movie-directory
 *       - t : slow motion in movie
 *       - u : turns on speed up in movie recording mode
 *       - i : increase speed up in movie recording
 ------------------------------------------------------------------------------- */
void key( unsigned char key, int, int )
{
    static int  pgon_mode = 0;
    static int  lighting = true;

    switch( key )
    {
        case 'q':
        case 27:
            delete mgr;
            osg::osgExit();
            exit( 0 );

        case ' ':
            pauseBench = !pauseBench;
            break;

        case 13:
            if( !pauseBench )
                pauseBench = true;
            else
                iterate();
            break;

        case 's':
            showing = !showing;
            if( showing )
            {
                beginEditCP( scene );
                scene->addChild( posNode );
                endEditCP( scene );
            }
            else
            {
                beginEditCP( scene );
                scene->subChild( posNode );
                endEditCP( scene );
            }
            break;

        case 'r':
            if( showing )
            {
                beginEditCP( scene );
                scene->subChild( posNode );
            }

            subRefCP( posNode );
            posNode = osg::Node::create();
            addRefCP( posNode );

            beginEditCP( posNode );
            posNode->setCore( osg::Group::create() );
            endEditCP( posNode );

            if( showing )
            {
                scene->addChild( posNode );
                endEditCP( scene );
            }
            break;

        case 'l':
            lighting = !lighting;
            if( lighting )
            {
                glDisable( GL_LIGHTING );
                puts( "lighting disabled" );
            }
            else
            {
                glEnable( GL_LIGHTING );
                puts( "lighting enabled" );
            }
            break;

        case 'p':
            pgon_mode++;
            if( pgon_mode > 2 )
                pgon_mode = 0;

            if( pgon_mode == 0 )
            {
                glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
                puts( "polygonMode: point" );
            }
            else if( pgon_mode == 1 )
            {
                glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
                puts( "polygonMode: line" );
            }
            else
            {
                glPolygonMode( GL_FRONT_AND_BACK, GL_POINT );
                puts( "polygonMode: fill" );
            }
            break;

        case 'a':
            finishedWinMode = false;
            break;

        case 'm':
            grabScreenOn = !grabScreenOn;
            break;

        case 't':
            slowMo = !slowMo;
            break;

        case 'u':
            speedUp = !speedUp;
            break;

        case 'i':
            speedUpFrames += 10;
            printf( "Speed Up: %d\n", speedUpFrames );
            break;

        default:
            printf( "key = %c %X\n", key, key );
    }
}

/** ----------------------------------------------------------------------------
 *  \ Main \
 ------------------------------------------------------------------------------- */
int main( int argc, char *argv[] )
{	
    osg::NodePtr    root;

    /* OSG init */
    osg::osgLog().setLogLevel( osg::LOG_WARNING );
    osg::osgInit( argc, argv );

    //loadDir( "modules" );
    loadDir( "." );

    
#ifndef _WIN32
	//Set locale format on Unix-like systems => float-separatpr "." 
	char *locale;
	locale = setlocale( LC_ALL, "en_IE.UTF-8");
	if (NULL == locale)
	{
		printf( "unable to set locale\n");
	}
#endif

    /* parse command line options */
    settings = new Settings( argc, argv );

    /* Background */
    if( settings->withWindow )
    {
        osg::GLUTWindowPtr  gwin;
        GLint               glvp[4];
	    int 		        winid;
        printf( "Setting Windows\n" );

        /* GLUT init */
        glutInit( &argc, argv );
	
        glutInitDisplayMode( GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE );
        glutInitWindowSize( 800, 800 );     /* before glutInit so user can */
        glutInitWindowPosition( 200, 200 ); /* override with comannd line args */
        winid = glutCreateWindow( "Collision Benchmark" );
        glutKeyboardFunc( key );
        glutVisibilityFunc( vis );
        glutReshapeFunc( reshape );
        glutDisplayFunc( display );
        glutMouseFunc( mouse );
        glutMotionFunc( motion );

        glutIdleFunc( display );

        glEnable( GL_DEPTH_TEST );
        glEnable( GL_LIGHTING );
        glEnable( GL_LIGHT0 );
        glLightModeli( GL_LIGHT_MODEL_TWO_SIDE, 1 );

        /* Window */
        glGetIntegerv( GL_VIEWPORT, glvp );
	    gwin = osg::GLUTWindow::create();
        gwin->setId( winid );
        gwin->init();
        gwin->setSize( glvp[2], glvp[3] );
        mgr = new osg::SimpleSceneManager;
        mgr->setWindow( gwin );
    }

    if( settings->method == METH_SPHERE )
        posFinder = new Sphere( settings );
    else if( settings->method == METH_CHECKSPHERE )
        posFinder = new CheckSphere( settings );
    else
        posFinder = new Grid( settings );

    if( settings->withWindow )
    {  
        osg::SimpleMaterialPtr  posNodeMat = osg::SimpleMaterial::create();
        scene = posFinder->getScene();
        mgr->setRoot( scene );
        mgr->setHeadlight( true );

        if( settings->whiteBackground )
        {
            osg::SolidBackgroundPtr bg = osg::SolidBackground::create();
            osg::ViewportPtr        vp = mgr->getWindow()->getPort( 0 );
            beginEditCP( bg );
            bg->setColor( osg::Color3f( 1, 1, 1 ) );
            endEditCP( bg );

            beginEditCP( vp );
            vp->setBackground( bg );
            endEditCP( vp );
        }

        grabScreenInit();

        posNode = osg::Node::create();
        addRefCP( posNode );
        dotNode = osg::makeSphereGeo( 1, 0.0075 );
        addRefCP( dotNode );

        beginEditCP( posNodeMat );
        posNodeMat->setDiffuse( osg::Color3f( 0.9, 0.0, 0.0 ) );
        posNodeMat->setAmbient( osg::Color3f( 0.9, 0.0, 0.0 ) );
        posNodeMat->setSpecular( osg::Color3f( .9, 0.0, 0.0 ) );
        endEditCP( posNodeMat );

        beginEditCP( dotNode );
        dotNode->setMaterial( posNodeMat );
        endEditCP( dotNode );

        beginEditCP( posNode );
        posNode->setCore( osg::Group::create() );
        endEditCP( posNode );
       
        
        if ( settings->useDistObj )
        {
                double dist = settings->dist;
                if ( dist < 0 ) 
                    dist = 0;    
                fixedNode = settings->getFixedNode();
                movingNode = settings->getMovingNode();
                settings->searchAlgorithm->api.shutdown();
                distNode = createDistanceObject( settings->useDistObj, fixedNode, dist );
                osg::NodePtr nodes[2] = { movingNode, distNode };
                settings->searchAlgorithm->api.prepareObjects( 2, nodes );

                distNodeMat = osg::SimpleMaterial::create();
                beginEditCP( distNodeMat );
                distNodeMat->setDiffuse( osg::Color3f( 0.0, 0.0, 0.9 ) );
                distNodeMat->setAmbient( osg::Color3f( 0.0, 0.0, 0.9 ) );
                distNodeMat->setSpecular( osg::Color3f( 0.0, 0.0, 0.9 ) );
                distNodeMat->setTransparency( 0.95 ); 
                endEditCP( distNodeMat );

                osg::GeometryPtr distNodeGeo = getGeom( distNode );
                beginEditCP( distNodeGeo );
                distNodeGeo->setMaterial( distNodeMat );
                endEditCP( distNodeGeo );

                beginEditCP( scene );
                scene->addChild( distNode );
                endEditCP( scene );
        }
        
        mgr->showAll();

        glutMainLoop();
    }
    else
    {
        settings->printHeader();
        posFinder->testAll();
        posFinder->printBottom();
    }

    return 0;
}

/* -----------------------------------------------------------------------------
 *  \brief Init screen grabbing
 ------------------------------------------------------------------------------- */
void grabScreenInit()
{
    osg::ViewportPtr    vp = mgr->getWindow()->getPort( 0 );

    /* create the foreground for screenshot functionality */
    fileGrab = osg::FileGrabForeground::create();
    beginEditCP( fileGrab );
    fileGrab->setActive( false );
    fileGrab->setName( "./movies/screenshot%04d.png" );
    endEditCP( fileGrab );

    /* add this one to the viewport, too */
    beginEditCP( vp );
    vp->getMFForegrounds()->push_back( fileGrab );
    endEditCP( vp );
}

/* -----------------------------------------------------------------------------
 *  \brief grab a screen shot into ./movies directory
 ------------------------------------------------------------------------------- */
void grabScreen()
{

    /* create the foreground for screenshot functionality */
    beginEditCP( fileGrab );
    fileGrab->setActive( true );
    fileGrab->setName( "./movies/screenshot%04d.png" );
    endEditCP( fileGrab );
}
