/*$6 ***************************************************************************
Copyright (C) <2007> <Sven Trenkel, Rene Weller, Gabriel Zachmann>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 ******************************************************************************* */
/**
 *  \brief The core system of the benchmarking suite: compute configurations
 *         and perform the benchmark
 *
 *  \author Sven Trenkel, Rene Weller, Gabriel Zachmann
 */
 
#include <OpenSG/OSGGeoFunctions.h>
#include <OpenSG/OSGSimpleGeometry.h>    
#include <OpenSG/OSGSolidBackground.h>
#include <OpenSG/OSGTriangleIterator.h>
#include <stdio.h>
#include "PosFinder.h"
#include "Utils.h"

using osg::     Pnt3f;
using osg::     Vec3f;

const double    EXACT = 0.005;

/* -----------------------------------------------------------------------------
 *  \brief load configurations from file
 ------------------------------------------------------------------------------- */
void PosFinder::loadQueue( const char *name )
{
    char        buf[65536];
    int         x = 0;
    FILE        *f;
    osg::Matrix m;

    f = fopen( name, "rt" );
    if( !f )
        return;

    fgets( buf, 65536, f );
    settings->dist = atof( buf );
    testQueue.resize( 0 );

    while( fgets( buf, 65536, f ) )
    {
        if( *buf == '\n' )
            continue;
        if( sscanf( buf, "%f %f %f %f", &m[x][0], &m[x][1], &m[x][2], &m[x][3]
            ) != 4 )
            break;
        if( ++x >= 4 )
        {
            x = 0;
            testQueue.push_back( m );
        }
    }

    fclose( f );
}

/* -----------------------------------------------------------------------------
 *  \brief test all configurations in settings
 ------------------------------------------------------------------------------- */
void PosFinder::testAll()
{
    const char  *File;

    while( (File = settings->getLoadname()) )
    {
        loadQueue( File );
        workQueue();
    }

    while( !settings->nextDist() )
    {
        if ( settings->useDistObj )
        {
            double dist = settings->dist;
            if ( dist < 0 ) dist = 0;
            settings->searchAlgorithm->api.shutdown();
            distNode = createDistanceObject( settings->useDistObj, fixedNode, dist );
            osg::NodePtr node[2] = { movingNode, distNode };
            settings->searchAlgorithm->api.prepareObjects( 2, node );
        }

        do
        {
            while( !iterate() );
        } while( !rotate() );
        if ( settings->useDistObj )
        {
            settings->searchAlgorithm->api.shutdown();
            osg::NodePtr node[2] = { movingNode, fixedNode };
            settings->searchAlgorithm->api.prepareObjects( 2, node );
        }
        workQueue();
    }
}

/* -----------------------------------------------------------------------------
 *  \brief Save all configurations in the configuration queue to a file
 *  \param name object name
 *  \param dist distance
 ------------------------------------------------------------------------------- */
void PosFinder::saveQueue( const char *name, double dist )
{
    char    FN[65536];
    FILE    *f;

    sprintf( FN, "%s_%f.pos\0", name, dist );
    f = fopen( FN, "wt" );
    if( !f )
    {
        fprintf( stderr, "Error writing file %s\n", FN );
        return;
    }

    fprintf( f, "%f\n\n", dist );
    for( int i = 0; i < testQueue.size(); ++i )
    {
        fprintf( f, "%f %f %f %f\n", testQueue[i][0][0], testQueue[i][0][1],
                 testQueue[i][0][2], testQueue[i][0][3] );
        fprintf( f, "%f %f %f %f\n", testQueue[i][1][0], testQueue[i][1][1],
                 testQueue[i][1][2], testQueue[i][1][3] );
        fprintf( f, "%f %f %f %f\n", testQueue[i][2][0], testQueue[i][2][1],
                 testQueue[i][2][2], testQueue[i][2][3] );
        fprintf( f, "%f %f %f %f\n\n", testQueue[i][3][0], testQueue[i][3][1],
                 testQueue[i][3][2], testQueue[i][3][3] );
    }

    fclose( f );
}

/* -----------------------------------------------------------------------------
 *  \brief Check all algorithms in configuration-queue
 *  \warning NanoTimer that is used for max/min-check time collision uses
 *   wall-clock time, not user time!
 *   Measurements with getTime() are exact with precision of 1 msec 
 ------------------------------------------------------------------------------- */
void PosFinder::workQueue()
{
    const Algorithm *algorithm;
    osg::NodePtr    prepNode[2] = { fixedNode, movingNode };

    if( settings->saveName )
        saveQueue( settings->saveName, settings->dist );
    printf( "%f", settings->dist );
    settings->resetAlgorithms();
    for( algorithm = settings->getAlgorithm(); algorithm;
         algorithm = settings->getAlgorithm() )
    {
        double  timecheck = 0.0; //for a single check
        double  mintimecheck = 1.7976931348623157e+308; //init minimum time for a single check with large value
        double  maxtimecheck = 0.0; //init maximum time for single check with small value
        double  start = getTime(); //start average timing
        int     colls = 0;
        for( int i = 0; i < testQueue.size(); ++i )
        {
            movingTrf->getSFMatrix()->setValue( testQueue[i] );
            nanotimer.start();
            if( algorithm->api.doCollisions( 1, &testQueue[i] ) )
            {
                ++colls;

                /*
                 * if (settings->dist >= 0) printf("WTF! Collision at (%f %f
                 * %f) with %s! Dist: %f\n", pos[0], pos[1], pos[2],
                 * algorithm->name, settings->getDist->api.getDistance(1, &m));
                 */
            }
            else if( settings->dist < 0 )
            {

                /*
                 * printf("WTF! No Collision at (%f %f %f) with %s!\n",
                 * pos[0], pos[1], pos[2], algorithm->name);
                 */
            }
            timecheck = nanotimer.elapsed();
            if( timecheck > maxtimecheck )
                maxtimecheck = timecheck;
            if( timecheck < mintimecheck )
                mintimecheck = timecheck;
        }
        
        double  dur = getTime() - start;

        printf( "\t" );
        if( !settings->noTiming )
        {
            printf( "%f", dur / testQueue.size() );
            
            if( settings->maxminTiming )
            {
                printf( "\t%f\t%f", mintimecheck, maxtimecheck );
            }
            
            if( settings->validate )
                printf( ":" );
        }

        if( settings->validate )
            printf( "%d/%u", colls, testQueue.size() );

        /*
         * printf( "%s braucht %f ms pro test und erkannte in %d von %u
         * Fllen eine Kollision\n", algorithm->name, dur / testQueue.size(),
         * colls, testQueue.size() );
         */
    }

    printf( "\n" );
    testQueue.resize( 0 );
}

/* -----------------------------------------------------------------------------
 *  \brief roatete moving object to next state
 ------------------------------------------------------------------------------- */
bool PosFinder::doRot()
{
    static const osg::Vec3f rotAxis[3] = { osg::Vec3f( 1, 0, 0 ), osg::Vec3f( 0, 1, 0 ), osg::Vec3f( 0, 0, 1 ) };

    double                  rot_rad[3] = { x_deg / 180.0 * 3.1415926538, y_deg / 180.0 * 3.1415926538, z_deg / 180.0 * 3.1415926538 };
    z_deg += settings->rotVel;
    if( z_deg >= 360 )
    {
        z_deg = 0;
        y_deg += settings->rotVel;
        if( y_deg >= 360 )
        {
            y_deg = 0;
            x_deg += settings->rotVel;
            if( x_deg > 180 )
            {
                x_deg = 0;
                matrix.setIdentity();
                movingTrf->getSFMatrix()->setValue( matrix );
                setSize();
                return true;
            }
        }
    }

    matrix = axisToMat( rotAxis[0], rot_rad[0] ) * axisToMat( rotAxis[1],
                                                              rot_rad[1] ) * axisToMat( rotAxis[2], rot_rad[2] );
    movingTrf->getSFMatrix()->setValue( matrix );
    setSize();
    return false;
}

/* -----------------------------------------------------------------------------
 *  \brief Add configuration matrix to the configuration-queue
 *         not: it is sufficient to change the translation, because the
 *         rotations are not affected by the configuration-finding algorithms
 ------------------------------------------------------------------------------- */
void PosFinder::queue( const osg::Pnt3f &pos )
{
    matrix.setTranslate( pos[0], pos[1], pos[2] );
    testQueue.push_back( matrix );
}

/* -----------------------------------------------------------------------------
 *  \brief Returns the distance between moving- and fixed object using
 *         distance computation of PQP or any other CD-library that offers
 *         functionality for distance computing
 *  \param pos translation of the moving object
 *  \return distance between moving- and fixed-object
 ------------------------------------------------------------------------------- */
double PosFinder::testPos( const osg::Pnt3f &pos )
{
    matrix.setTranslate( pos[0], pos[1], pos[2] );
    movingTrf->getSFMatrix()->setValue( matrix );

    return settings->searchAlgorithm->api.getDistance( 1, &matrix, NULL, NULL );
}

/* -----------------------------------------------------------------------------
 *  \brief Calculates
 *         distance computation of PQP or any other CD-library that offers
 *         functionality for distance computing
 *  \param pos translation of the moving object
 *  \return distance between moving- and fixed-object
 ------------------------------------------------------------------------------- */
bool PosFinder::testBetweenZero( const osg::Pnt3f    &pos1,
                                 const osg::Pnt3f    &pos2,
                                 osg::Pnt3f          &ret )
{
    bool        col1 = testBin( pos1 ), col2 = testBin( pos2 );
    osg::Pnt3f  min, max;

    osg::Vec3f  path;
    double      mind = 0, maxd = 1, av = 0.5;
    osg::Pnt3f  midPos;
    if( !col1 && col2 )
    {
        min = pos2;
        max = pos1;
    }
    else if( col1 && !col2 )
    {
        min = pos1;
        max = pos2;
    }
    else
    {
        return false;
    }

    path = max - min;
    midPos = min + path * 0.5;
    while( path.length() * (maxd - mind) > EXACT )
    {
        if( testBin( midPos ) )
            mind = av;
        else
            maxd = av;
        av = ( mind + maxd ) / 2;
        midPos = min + path * av;
    }

    if( settings->dist < 0 )
        ret = osg::Pnt3f( min + path * mind );
    else
        ret = osg::Pnt3f( min + path * maxd );
    return true;
}

/* -----------------------------------------------------------------------------
 *  \brief Checks if fixed- and moving-object overlap
 *  \param pos translation of the moving object
 *  \return true if objects overlap, false elsewhere
 ------------------------------------------------------------------------------- */
bool PosFinder::testBin( const osg::Pnt3f &pos )
{
    matrix.setTranslate( pos[0], pos[1], pos[2] );
    movingTrf->getSFMatrix()->setValue( matrix );

    return settings->searchAlgorithm->api.doCollisions( 1, &matrix );
}

/* -----------------------------------------------------------------------------
*  \brief Places moving object in settings->distance to the fixed object and adds
*         it to the configuration queue
*  \param pos1 translation of fixed object
*  \param pos2 translation of moving object
 ------------------------------------------------------------------------------- */
void PosFinder::queueBetween( const osg::Pnt3f &pos1, const osg::Pnt3f &pos2 )
{
    osg::Pnt3f ret( 0, 0, 0 );

    if( testBetween( pos1, pos2, ret ) )
        queue( ret );
}

/* -----------------------------------------------------------------------------
*  \brief Places moving object in settings->distance to the fixed object by
*         doing a binary search
*  \param pos1 translation of fixed object
*  \param pos2 translation of moving object
*  \param ret  translation with required distance
*  \return true, if it is possible to place the object correctly
 ------------------------------------------------------------------------------- */
bool PosFinder::testBetween( const osg::Pnt3f    &pos1,
                             const osg::Pnt3f    &pos2,
                             osg::Pnt3f          &ret )
{
    double      dist1, dist2;
    osg::Pnt3f  min, max;
    osg::Pnt3f  midPos;
    double      mind = 0, maxd = 1, av = 0.5, maxdist;
    osg::Vec3f  path;

    if( settings->dist <= 0 || settings->useDistObj )
    {
        return testBetweenZero( pos1, pos2, ret );
    }

    dist1 = testPos( pos1 );
    dist2 = testPos( pos2 );
    if( dist1 >= settings->dist && dist2 < settings->dist )
    {
        min = pos2;
        max = pos1;
    }
    else if( dist1 < settings->dist && dist2 >= settings->dist )
    {
        min = pos1;
        max = pos2;
    }
    else
    {
        return false;
    }
    
    path = max - min;
    midPos = min + path * 0.5;
    maxdist = dist2;
    
    do
    {
        double  dist = testPos( midPos );
        if( dist > settings->dist )
        {
            maxd = av;
            maxdist = dist;
        }
        else
        {
            mind = av;
        }

        av = ( mind + maxd ) / 2;
        midPos = min + path * av;
    } while( maxdist - settings->dist > EXACT );

    ret = osg::Pnt3f( min + path * maxd );
    return true;
}

/* -----------------------------------------------------------------------------
 *  \brief Create the root node of the scene and add the moving- and the fixed
 *         object
 ------------------------------------------------------------------------------- */
osg::NodePtr PosFinder::getScene()
{
    if( scene != osg::NullFC )
        return scene;

    scene = osg::Node::create();

    beginEditCP( scene );
    scene->setCore( osg::Group::create() );
    scene->addChild( fixedNode );
    scene->addChild( trfNode );
    endEditCP( scene );

    return scene;
}

/* -----------------------------------------------------------------------------
 *  \brief calculates the sizes of moving- and fixed-object
 ------------------------------------------------------------------------------- */
void PosFinder::setSize()
{
    osg::Vec3f          fixedMin, fixedMax, movingMin, movingMax;

    osg::DynamicVolume  vol = movingNode->getVolume( true );
    fixedNode->getVolume( true ).getBounds( fixedMin, fixedMax );
    vol.transform( matrix );
    vol.getBounds( movingMin, movingMax );
    for( int i = 0; i < 3; ++i )
    {
        fixedCenter[i] = ( fixedMin[i] + fixedMax[i] ) / 2;
        movingCenter[i] = ( movingMin[i] + movingMax[i] ) / 2;
    }

    fixedSize = fixedMax - fixedCenter;
    movingSize = movingMax - movingCenter;
}

/* -----------------------------------------------------------------------------
 *  \brief Initialization of the position finder: Setting up parameters
 *         setting up the algorithms, set sizes of the objects,...
 ------------------------------------------------------------------------------- */
PosFinder::PosFinder( Settings *set ) : settings( set ), scene( osg::NullFC ), matrix(), x_deg( 0 ), y_deg( 0 ), z_deg( 0 ), nanotimer()
{
    osg::GeometryPtr    fixedGeom, movingGeom;

    if( !settings->searchAlgorithm->api.getDistance )
    {
        if ( !settings->useDistObj )
        {
            fprintf( stderr, "Error: %s(%s) does not support distance computation. Choose an algorithm that does (-F) or use distance objects (-D)\n", settings->searchAlgorithm->shortName, settings->searchAlgorithm->name );
            exit( 1 );
        }
        else
        {
            fprintf( stderr, "Useing %s(%s) for position searching with distance objects.\n", settings->searchAlgorithm->shortName, settings->searchAlgorithm->name );
        }
    }
    else
    {
        if ( !settings->useDistObj )
        {
            fprintf( stderr, "Useing %s(%s) for distance computation\n", settings->searchAlgorithm->shortName, settings->searchAlgorithm->name );
        }
        else
        {
            fprintf( stderr, "Warning: Useing %s(%s) for position searching with distance objects. This is slow and unreliable.\n", settings->searchAlgorithm->shortName, settings->searchAlgorithm->name );
        }
    }

    fixedNode = settings->getObject();
    fixedGeom = getGeom( fixedNode );

    movingNode = settings->getObject();
    movingGeom = getGeom( movingNode );
    
    osg::NodePtr        node[2] = { movingNode, fixedNode };
    bool                gotDist = false;


    matrix.setIdentity();

    /* transformation for moving node */
    trfNode = osg::Node::create();
    movingTrf = osg::Transform::create();
    beginEditCP( trfNode );
    trfNode->setCore( movingTrf );
    trfNode->addChild( movingNode );
    endEditCP( trfNode );

    settings->resetAlgorithms();
    double  start = getTime(); //start build timing
    
    for( const Algorithm * algo = settings->getAlgorithm(); algo;
         algo = settings->getAlgorithm() )
    {
        if (settings->useDistObj && algo == settings->searchAlgorithm ) continue;
        algo->api.prepareObjects( 2, node );    
        
        if( algo == settings->searchAlgorithm )
            gotDist = true;
    }
    
    double end = getTime(); //stop build time
    buildTime = end-start;

    if( !gotDist )
        settings->searchAlgorithm->api.prepareObjects( 2, node );

    setSize();
}

/*! ----------------------------------------------------------------------------
 *  \brief  Display build time and #faces at the command line
 ------------------------------------------------------------------------------- */
void PosFinder::printBottom()
{
    double avArea = 0;
    unsigned int numfacesmoving = countTriangles( getGeom( movingNode ) , &avArea );
    unsigned int numfacesfixed = countTriangles( getGeom( fixedNode ) , &avArea );
    if ( numfacesmoving != numfacesfixed )
    {
        printf( "# num faces moving = %d\n", numfacesmoving );
        printf( "# num faces fixed = %d\n", numfacesfixed );
    }
    else
        printf( "# num faces = %d\n", numfacesmoving );
        
    printf( "# build time = %f (msec)\n", buildTime );
}
